;inc_rand        = 1    ;1 is the best option and default
;#define inc_seedrand   ;If defined, this includes seedrand
;#define inc_randrange8 ;If defined, this includes randrange8
;#define inc_randrange16;If defined, this includes randrange16
;#define inc_rand10     ;If defined, this includes rand10
;#define inc_randTI     ;If defined, this includes randTI
;#define inc_HL_Times_A ;If defined, this includes HL_Times_A
;#define inc_DE_Times_A ;If defined, this includes DE_Times_A
;#define inc_HL_Times_BC;If defined, this includes HL_Times_BC
;#define inc_DE_Times_BC;If defined, this includes DE_Times_BC
;#define smc            ;If defined, this uses SMC where applicable. If possible, use SMC.


;;random.inc
;;Random Number Generator Routines
;;Author:
;;    Zeda Thomas (xedaelnara@gmail.com)
;;License:
;;    +All parts of this document may be used for any purpose, but must remain open source and freely available.
;;      ++This code should not be goverened by any other license,
;;      ++BUT this code may be used in open or closed source projects, with the stipulation:
;;    +Give some credit to the author, even if only mentioned in the source code.
;;      ++This is mainly so that questions and issues can be reported to me, and I am genuinely interested when my code is used!
;;      ++If this document is made available with source, and this license info is unchanged and included in this file, that will suffice as "credit."
;;      ++Also forgot to mention fame and glory, hail hydra.
;;Disclaimer:
;;    +These are NOT cryptographically secure. These are quality PRNGs for games and simulation.
;;Routines:
;;  routine         in         out              size    speed
;;=================================================================
;;  rand  =11       ---         HL              
;;  rand  =10       ---         HL              
;;  rand  =9        ---         HL              14      75
;;  rand  =8        ---         HL              huge    too long+32 if no smc  
;;  rand  =7        ---         HL              18      87      +6 if no SMC  
;;  rand  =6        ---         HL              14      58|60   +6 if no SMC  
;;  rand  =5        ---         HL              44      219     +24 if no SMC  
;;  rand  =4        ---         HL              26      148     +12 if no SMC  
;;  rand  =3        ---         HL              50      292~294
;;  rand  =2        ---         HL              106     606
;;  rand  =1        ---         HL              58      291     +32 if no SMC
;;  randrange8      B           A in [0,B-1]    5       249~342 +rand
;;  randrange16     BC          HL in [0,BC-1]  26      571~757 +rand
;;  rand10          ---         A in [0,9]      16      102     +rand
;;  randTI          HL          (hl)            46      1652    +7*rand
;;  HL_Times_A      HL,A        AHL             52      211~304
;;  DE_Times_A      DE,A        AHL             51      207~300
;;  HL_Times_BC     HL,BC       HLDE            21      529~715
;;  DE_Times_BC     DE,BC       HLDE            20      525~711
;;=================================================================
;;                  score    similar, but better
;;custom      = 0
;;rand_best   = 1   *******
;;lcg_lfsr    = 2   *****   Use [1] instead.
;;lehmer_lfsr = 3   ***     Use [4] instead.
;;rand_Axe    = 4   ******  11 hours, 46 minutes, 16 seconds to cycle 
;;lcg_lfsr24  = 5   ******  Compromise of [1] and [4], 130+ years at 15MHz to cycle.
;;rand_cadan8 = 6   *
;;rand_cadan16= 7   ***
;;rand_mlcg   = 8   ***
;;rand_Ion    = 9   ****
#ifndef inc_rand
    #define inc_rand rand_best
    #endif
#ifdef inc_randrange8
    #define inc_HL_Times_A
    #endif
#ifdef inc_randrange16
    #define inc_HL_Times_A
    #endif
;;%=%=%=%=%=%=%=%=%=%=%=%=%=%=%=%=%=%=%=%=%=%=%=%=%=%=%=%=%=%=%=%=%
rand:
#if inc_rand==11
#endif
#if inc_rand==10
#endif
#if inc_rand==9
;from ionRandom
#ifdef smc
seed=$+1
    ld hl,0
#else
    ld hl,(seed)
#endif
    ld a,r
    ld d,a
    ld e,(hl)
    add hl,de
    add a,l
    xor h
    ld (seed),hl
    ret
#endif
#if inc_rand==8
;Inputs: seed1,seed2
;Outputs:
; HLA is the pseudo-random number
; seed1,seed2 incremented accordingly
;Destroys: BC,DE
;Notes:
; seed1*243+83 mod 65519 -> seed1
; seed2*251+43 mod 65521 -> seed2
; Output seed1*seed2 mod 16777259
#ifdef smc
seed1=$+1
    ld hl,13579
#else
    ld hl,(seed1)
#endif
    xor a
    ld b,h \ ld c,l
    ld de,83
    add hl,hl \ rla    ;2
    add hl,bc \ adc a,d    ;3
    add hl,hl \ rla    ;6
    add hl,bc \ adc a,d    ;7
    add hl,hl \ rla    ;14
    add hl,bc \ adc a,d    ;15
    add hl,hl \ rla    ;30
    add hl,hl \ rla    ;60
    add hl,hl \ rla    ;120
    add hl,bc \ adc a,d    ;121
    add hl,hl \ rla    ;242
    add hl,bc \ adc a,d    ;243
    add hl,de \ adc a,d    ;243x+83
;now AHL mod 65519
; Essentially, we can at least subtract A*65519=A(65536-17), so add A*17 to HL
    ex de,hl
    ld l,a
    ld b,h
    ld c,l
    add hl,hl
    add hl,hl
    add hl,hl
    add hl,hl
    add hl,bc
    add hl,de
    ld c,17
    jr nc,$+3
    add hl,bc
    add hl,bc
    jr c,$+4
    sbc hl,bc
    ld (seed1),hl
;seed1 done, now we need to do seed2
#ifdef smc
seed2=$+1
    ld hl,13579
#else
    ld hl,(seed2)
#endif
; seed1*243+83 mod 65519 -> seed1
; seed2*251+43 mod 65521 -> seed2
; Output seed1*seed2
    xor a
    ld b,h \ ld c,l
    ld de,43
    add hl,hl \ rla  ;2
    add hl,bc \ adc a,d ;3
    add hl,hl \ rla  ;6
    add hl,bc \ adc a,d ;7
    add hl,hl \ rla  ;14
    add hl,bc \ adc a,d ;15
    add hl,hl \ rla  ;30
    add hl,bc \ adc a,d ;31
    add hl,hl \ rla  ;62
    add hl,hl \ rla  ;124
    add hl,bc \ adc a,d ;125
    add hl,hl \ rla  ;250
    add hl,bc \ adc a,d ;251
    add hl,de \ adc a,d ;251x+83
;now AHL mod 65521
; Essentially, we can at least subtract A*65521=A(65536-15), so add A*15 to HL
    ex de,hl
    ld l,a
    ld b,h
    ld c,l
    add hl,hl
    add hl,hl
    add hl,hl
    add hl,hl
    sbc hl,bc
    add hl,de
    ld c,15
    jr nc,$+3
    add hl,bc
    add hl,bc
    jr c,$+4
    sbc hl,bc
    ld (seed2),hl
;now seed1 and seed 2 are computed
    ld bc,(seed1)
    ex de,hl
    call BC_Times_DE
    ex de,hl
    ld l,b
    ld h,0
    ld b,h
    ld c,l
    add hl,hl
    add hl,hl
    add hl,bc
    add hl,hl
    add hl,hl
    add hl,bc
    add hl,hl
    add hl,bc
    ld c,d
    ld d,e
    ld e,a
    ld a,c
    sbc hl,bc
    sbc a,b
    ret nc
    ld c,43
    add hl,bc
    ret
BC_Times_DE:
;BHLA is the result
 ld a,b
 or a
 ld hl,0
 ld b,h
;1
 add a,a
 jr nc,$+4
 ld h,d
 ld l,e
;2
 add hl,hl
 rla
 jr nc,$+4
 add hl,de
 adc a,b
;227+10b-7p
 add hl,hl
 rla
 jr nc,$+4
 add hl,de
 adc a,b

 add hl,hl
 rla
 jr nc,$+4
 add hl,de
 adc a,b

 add hl,hl
 rla
 jr nc,$+4
 add hl,de
 adc a,b

 add hl,hl
 rla
 jr nc,$+4
 add hl,de
 adc a,b

 add hl,hl
 rla
 jr nc,$+4
 add hl,de
 adc a,b

 add hl,hl
 rla
 jr nc,$+4
 add hl,de
 adc a,b

;===
;AHL is the result of B*DE*256
 push hl
 ld h,b
 ld l,b
 ld b,a
 ld a,c
 ld c,h
;1
 add a,a
 jr nc,$+4
 ld h,d
 ld l,e
;2
 add hl,hl
 rla
 jr nc,$+4
 add hl,de
 adc a,c
;227+10b-7p
 add hl,hl
 rla
 jr nc,$+4
 add hl,de
 adc a,c

 add hl,hl
 rla
 jr nc,$+4
 add hl,de
 adc a,c

 add hl,hl
 rla
 jr nc,$+4
 add hl,de
 adc a,c

 add hl,hl
 rla
 jr nc,$+4
 add hl,de
 adc a,c

 add hl,hl
 rla
 jr nc,$+4
 add hl,de
 adc a,c

 add hl,hl
 rla
 jr nc,$+4
 add hl,de
 adc a,c

 pop de
;Now BDE*256+AHL
 ld c,a
 ld a,l
 ld l,h
 ld h,c
 add hl,de
 ret nc
 inc b
;BHLA is the 32-bit result
 ret
#endif
#if inc_rand==7
;  From CaDan, by Iambian and Geekboy, modified for this source
;  Output is 16 bit, cycle is 65535
#ifdef smc
seed=$+1
    ld hl,9999
#else
    ld hl,(seed)
#endif
    ld a,l
    rlca
    xor h
    rlca  ;
    rlca  ;
    xor h ;
    rlca
    xor h
    rlca
    adc hl,hl
    ld (seed),hl
    ret
#endif
#if inc_rand==6
;  From CaDan, by Iambian and Geekboy, modified for this source
;  Output is only 8 bit, cycle is 255
#ifdef smc
seed=$+1
    ld a,99
#else
    ld a,(seed)
#endif
    add a,a
    jr nc,$+4
    xor %00011101
    ld h,a
    ld (seed),a
    cpl
    ld l,a
    ret
#endif
#if inc_rand==5
;rand24:
;;219cc
;;44 bytes
#ifdef smc
seed1_0=$+1
    ld hl,12345
seed1_1=$+1
    ld a,67
#else
    ld hl,(seed1_0)
    ld a,(seed1_1)
#endif
    ld b,h
    ld c,l
    ld d,a
    add hl,hl \ rla
    add hl,hl \ rla
    inc l
    add hl,bc \ adc a,0
    ld (seed1_0),hl
    ld (seed1_1),a
    ld c,b
    ld b,a
#ifdef smc
seed2_0=$+1
    ld hl,65432
seed2_1=$+1
    ld a,10
#else
    ld hl,(seed2_0)
    ld a,(seed2_1)
#endif
    add hl,hl
    rla
    ld (seed2_1),a
    sbc a,a
    and %10000111
    xor l
    ld l,a
    ld (seed2_0),hl
    add hl,bc
    ret
#ifdef inc_seedrand
seedrand:
    ld a,r
    ld h,a
    cpl
    ld l,a
    ld (seed1_0),hl
    ld (seed2_0),hl
    ret
#endif
#endif
#if inc_rand==rand_Axe
;;same algo as used in Axe
;;148cc
;;26 bytes
#ifdef smc
seed1=$+1
    ld hl,12345
#else
    ld hl,(seed1)
#endif
    ld b,h
    ld c,l
    add hl,hl
    add hl,hl
    inc l
    add hl,bc
    ld (seed1),hl
#IFDEF smc
seed2=$+1
    ld hl,6789
#ELSE
    ld hl,(seed2)
#ENDIF
    add hl,hl
    sbc a,a
    and %00101101
    xor l
    ld l,a
    ld (seed2),hl
    add hl,bc
    ret  
#ifdef inc_seedrand
seedrand:
    ld a,r
    ld h,a
    cpl
    ld l,a
    ld (seed1),hl
    ld (seed2),hl
    ret
#endif
#endif
#if inc_rand==1
;rand32:
;;Tested and passes all CAcert tests
;;Uses a very simple 32-bit LCG and 32-bit LFSR
;;it has a period of 18,446,744,069,414,584,320
;;roughly 18.4 quintillion.
;;LFSR taps: 0,2,6,7  = 11000101
;;291cc  +32cc if no SMC
#ifdef smc
seed1_0=$+1
    ld hl,12345
seed1_1=$+1
    ld de,6789
#else
    ld hl,(seed2_0)
    ld de,(seed2_1)
#endif
    ld b,h
    ld c,l
    add hl,hl \ rl e \ rl d
    add hl,hl \ rl e \ rl d
    inc l
    add hl,bc
    ld (seed1_0),hl
    ld hl,(seed1_1)
    adc hl,de
    ld (seed1_1),hl
    ex de,hl
;;lfsr
#ifdef smc
seed2_0=$+1
    ld hl,9876
seed2_1=$+1
    ld bc,54321
#else
    ld hl,(seed2_0)
    ld bc,(seed2_1)
#endif
    add hl,hl \ rl c \ rl b
    ld (seed2_1),bc
    sbc a,a
    and %11000101
    xor l
    ld l,a
    ld (seed2_0),hl
    ex de,hl
    add hl,bc
    ret
#ifdef inc_seedrand
seedrand:
    ld a,r
    ld h,a
    cpl
    ld l,a
    ld (seed1_0),hl
    ld (seed2_1),hl
    ret
#endif
#endif
#IF inc_rand==2
;zrng:
;;Inputs: (seed1), (seed2)
;;Outputs: HL
;;Notes:
;;  I submitted this to the CAcert research lab and the results:
;;      Data Size   Entropy (->8)   Birthday Spacing    Matrix Ranks    6x8 Matrix Rank     Minimum Distance Test   Random Spheres Test  The Sqeeze Test  Overlapping Sums Test
;;      16 MiB      7.999990        0.877253            0.303           0.431               0.904194                0.842481             0.664840         0.001861
;;      Pass        Pass            Pass                Pass            Pass                Pass                    Pass                 Pass             Inconclusive
;;  One test came back inconclusive, the rest passed.
;;  That said, this is NOT cryptographically secure. With knowledge of how the output is formed, it takes only observing a short sequence of outputs (maybe 16?) to correctly guess the hidden parameters.
;;Speed: 606cc
;;Size: 106 bytes
#ifdef smc
msw=$+1
    ld bc,12345
lsw=$+1
    ld de,6789
#else
    ld bc,(msw)
    ld de,(lsw)
#endif
    ld h,c
    ld l,d
    push hl
    ld h,e
    xor a
    ld l,a
    sbc hl,de \ ex(sp),hl \ sbc hl,bc
    or a \ ex(sp),hl \ sbc hl,de \ ex(sp),hl \ sbc hl,bc
    or a \ ex(sp),hl \ sbc hl,de \ ex de,hl \ pop hl \ sbc hl,bc
    ld b,h \ ld c,l
    ld h,l \ ld l,d \ ld ixh,e \ ld ixl,a
    add ix,de \ adc hl,bc \ ex de,hl \ ld a,c
    add hl,hl \ rla \ rl c
    add hl,hl \ rla \ rl c
    add hl,hl \ rla \ rl c
    add hl,hl \ rla \ rl c
    ld c,a
    ex de,hl
    inc e
    add ix,de
    adc hl,bc
    ld (msw),hl
    ld (lsw),ix
    ex de,hl
#ifdef smc
msw2=$+1
    ld bc,9867
lsw2=$+1
    ld hl,54321
#else
    ld bc,(msw2)
    ld hl,(lsw2)
#endif
    add hl,hl
    rl c
    rl b
    sbc a,a
    and %01000111
    xor l
    ld l,a
    ld (lsw2),hl
    ld (msw2),bc
    add hl,de
    ret
#ifdef inc_seedrand
seedrand:
    ld a,r
    ld h,a
    cpl
    ld l,a
    ld (lsw),hl
    ld (msw2),hl
    ret
#endif
#ENDIF
#IF inc_rand==3
;;Input:
;;  (seed1) is a non-zero 16-bit int
;;  (seed2) is a 16-bit int
;;Output:
;;  HL is the pseudo-random number
;;  DE is the output of the LFSR
;;  BC is the previous seed of the Lehmer PRNG
;;  (seed1) is the output of the LFSR
;;  (seed2) is the output of the Lehmer PRNG
;;Destroys: A
;;Notes:
;;  This uses a 16-bit Lehmer PRNG, and an LFSR
;;  The period is 4,294,901,760 (65536*65535)
;;  Technically,the Lehmer PRNG here can have values from 1 to 65536. In this application, we store 65536 as 0.
;;Speed:
;;    If smc = 0, add 12cc
;;    293+a-c, a,c are {0,1}
;;      probability a= 1 is 38/65536
;;      probability c= 1 is 1/65536
;;    Average: 293+37/65536 = 293.0005645751953125cc
;;    Best:292cc
;;    Worst:294cc
;;50 bytes
#IFDEF smc
seed2 = $+1
    ld hl,0
#ELSE
    ld hl,(seed2)
#ENDIF
;multiply by 75
    ld c,l
    ld b,h
    xor a
    adc hl,hl \ jr nz,$+3 \ inc a \ rla
    add hl,hl \ rla
    add hl,hl \ rla \ add hl,bc \ adc a,d
    add hl,hl \ rla
    add hl,hl \ rla \ add hl,bc \ adc a,d
    add hl,hl \ rla \ add hl,bc
    ld e,a
    sbc hl,de
    jr nc,$+3
    inc hl
    ld (seed2),hl
#IFDEF smc
seed1 = $+1
    ld hl,1
#ELSE
    ld hl,(seed1)
#ENDIF
    add hl,hl
    sbc a,a
    and %00101101
    xor l
    ld l,a
    ld (seed1),hl
    add hl,bc
    ret
#ifdef inc_seedrand
seedrand:
    ld a,r
    ld h,a
    cpl
    ld l,a
    ld (seed1),hl
    ld (seed2),hl
    ret
#endif
#ENDIF
#IFDEF inc_randrange8
randrange8:
;;Input: B
;;Output: A is a number on [0,B-1]
;;Example, if B is 10, this outputs an int from 0 to 9
;;562cc~660cc
    push bc
    call rand
    pop af
#ENDIF
#ifdef inc_HL_Times_A
HL_Times_A:
    ex de,hl
#define inc_DE_Times_A
#endif
#ifdef inc_DE_Times_A
DE_Times_A:
;;51 bytes
;;300cc worst, 207 best
    ld hl,0 \ ld c,l
    add a,a \ jr nc,$+5 \ ld h,d \ ld l,e
    add hl,hl \ rla \ jr nc,$+4 \ add hl,de \ adc a,c
    add hl,hl \ rla \ jr nc,$+4 \ add hl,de \ adc a,c
    add hl,hl \ rla \ jr nc,$+4 \ add hl,de \ adc a,c
    add hl,hl \ rla \ jr nc,$+4 \ add hl,de \ adc a,c
    add hl,hl \ rla \ jr nc,$+4 \ add hl,de \ adc a,c
    add hl,hl \ rla \ jr nc,$+4 \ add hl,de \ adc a,c
    add hl,hl \ rla \ ret nc \ add hl,de \ adc a,c
    ret
#endif
#IFDEF inc_randrange16
randrange16:
;;->BC   ->HL
    push bc
    call rand
    pop bc
#define inc_HL_Times_BC
#endif
#ifdef inc_HL_Times_BC
HL_Times_BC:
    ex de,hl
#define inc_DE_Times_BC
#endif
#ifdef inc_DE_Times_BC
DE_Times_BC:
    ld a,c
    call DE_Times_A
    push hl
    ld c,b
    ld b,a
    ld a,c
    call DE_Times_A
    pop de
    ld c,d
    add hl,bc
    adc a,0
    ld d,l
    ld l,h
    ld h,a
    ret
#endif
#IFDEF inc_randTI
randTI:
;;Inputs: HL points to where the float should be written (does not allocate RAM).
;;Outputs: Stores a random floating point number on [0,1) to where HL points
;;1652cc+rand
;;46 bytes
    ld (hl),0 \ inc hl
    ld (hl),7Fh
    ld b,7
randTIloop:
    push bc
    push hl
    call rand
    xor a
    ld b,h
    ld c,l
    ld d,a
    add hl,hl \ rla
    add hl,hl \ rla
    add hl,bc \ adc a,d
    add hl,hl \ rla
    ld e,a
    xor a
    add hl,hl \ rla
    add hl,hl \ rla
    add hl,bc \ adc a,d
    add hl,hl \ rla
    rlca
    rlca
    rlca
    rlca
    or e
    pop hl
    inc hl
    ld (hl),a
    pop bc
    djnz randTIloop
    ret
#ENDIF
#IFDEF inc_rand10
rand10:
;;Output: A is an int from 0 to 9.
;;102cc+rand
;;16 bytes
    call rand
    xor a
    ld b,h
    ld c,l
    add hl,hl \ rla
    add hl,hl \ rla
    add hl,bc \ adc a,0
    add hl,hl \ rla
    ret
#ENDIF


.echo "random.inc :",$-rand," bytes"
